from collections import defaultdict

from wikidataintegrator import wdi_core
from wikidataintegrator.wdi_helpers import id_mapper


class WikibaseHelper:
    """
    Helper functions for accessing PIDs and QIDs across wikibases

    This functionality depends upon their existing two properties: equivalent property and equivalent class
    with the equivalent property values: 'http://www.w3.org/2002/07/owl#equivalentProperty' and
    'http://www.w3.org/2002/07/owl#equivalentClass' respectively
    """

    def __init__(self, sparql_endpoint_url='https://query.wikidata.org/sparql'):
        self.sparql_endpoint_url = sparql_endpoint_url
        # a map of property URIs to a PID in the wikibase you are using
        try:
            equiv_prop_pid = self.guess_equivalent_property_pid()
        except Exception:
            raise ValueError("Error: No property found with URI 'http://www.w3.org/2002/07/owl#equivalentProperty'")
        uri_pid = id_mapper(equiv_prop_pid, endpoint=self.sparql_endpoint_url, return_as_set=True)
        # remove duplicates/conflicts
        self.URI_PID = {k: list(v)[0] for k, v in uri_pid.items() if len(v) == 1}
        # get equivalent class PID
        if 'http://www.w3.org/2002/07/owl#equivalentClass' not in self.URI_PID:
            raise ValueError("Error: No property found with URI 'http://www.w3.org/2002/07/owl#equivalentClass'")
        equiv_class_pid = self.URI_PID['http://www.w3.org/2002/07/owl#equivalentClass']
        # a map of item URIs to a QID in the wikibase you are using
        uri_qid = id_mapper(equiv_class_pid, endpoint=self.sparql_endpoint_url,
                            return_as_set=True)
        # remove duplicates/conflicts
        self.URI_QID = {k: list(v)[0] for k, v in uri_qid.items() if len(v) == 1}

    def guess_equivalent_property_pid(self):
        # get the equivalent property PID without knowing the PID for equivalent property!!!
        query = '''SELECT * WHERE {
          ?item ?prop <http://www.w3.org/2002/07/owl#equivalentProperty> .
          ?item <http://wikiba.se/ontology#directClaim> ?prop .
        }'''
        pid = wdi_core.WDItemEngine.execute_sparql_query(query, endpoint=self.sparql_endpoint_url)
        pid = pid['results']['bindings'][0]['prop']['value']
        pid = pid.split("/")[-1]
        return pid

    def get_pid(self, uri):
        """
        Get the pid for the property in this wikibase instance ( the one at `sparql_endpoint_url` ),
         that corresponds to (i.e. has the equivalent property) `uri`
        """
        # if the wikibase is wikidata, and we give a wikidata uri or a PID with no URI specified:
        if (self.sparql_endpoint_url == 'https://query.wikidata.org/sparql' and
                (uri.startswith("P") or uri.startswith("http://www.wikidata.org/entity/"))):
            # don't look up anything, just return the same value
            return uri
        if uri.startswith("P"):
            uri = "http://www.wikidata.org/entity/" + uri
        return self.URI_PID[uri]

    def get_qid(self, uri):
        # if the wikibase is wikidata, and we give a wikidata uri or a QID with no URI specified:
        if (self.sparql_endpoint_url == 'https://query.wikidata.org/sparql' and
                (uri.startswith("Q") or uri.startswith("http://www.wikidata.org/entity/"))):
            # don't look up anything, just return the same value
            return uri
        if uri.startswith("Q"):
            uri = "http://www.wikidata.org/entity/" + uri
        return self.URI_QID[uri]

    def prop2qid(self, prop, value):
        """
        Lookup the local item QID for a Wikidata item that has a certain `prop` -> `value`
        in the case where the local item has a `equivalent item` statement to that wikidata item
        Example: In my wikibase, I have CDK2 (Q79363) with the only statement:
            equivalent class -> http://www.wikidata.org/entity/Q14911732
            Calling prop2qid("P351", "1017") will return the local QID (Q79363)
        :param prop:
        :param value:
        :return:
        """
        equiv_class_pid = self.URI_PID['http://www.w3.org/2002/07/owl#equivalentClass']
        query = """
        PREFIX wwdt: <http://www.wikidata.org/prop/direct/>
        SELECT ?wditem ?localitem ?id WHERE {{
          SERVICE <https://query.wikidata.org/sparql> {{
            ?wditem wwdt:{prop} "{value}"
          }}
          ?localitem wdt:{equiv_class_pid} ?wditem
        }}"""
        query = query.format(prop=prop, value=value, equiv_class_pid=equiv_class_pid)

        results = wdi_core.WDItemEngine.execute_sparql_query(query, endpoint=self.sparql_endpoint_url)
        result = results['results']['bindings']
        if len(result) == 0:
            return None
        elif len(result) > 1:
            raise ValueError("More than one wikidata ID found for {} {}: {}".format(prop, value, result))
        else:
            return result[0]['localitem']['value'].split("/")[-1]

    def id_mapper(self, prop, filters=None, return_as_set=False):
        """
        # see wdi_helpers.id_mapper for help on usage
        # QIDs returned are from the wikibase specified in self.sparql_endpoint_url
        # see WikibaseHelper.prop2qid for more details
        Example: Get all human genes:
        prop = "P351"
        filters = [("P703", "Q15978631")]
        h.id_mapper(prop, filters)
        """
        if filters:
            filter_str = "\n".join("?wditem wwdt:{} wwd:{} .".format(x[0], x[1]) for x in filters)
        else:
            filter_str = ""
        query = """
        PREFIX wwdt: <http://www.wikidata.org/prop/direct/>
        PREFIX wwd: <http://www.wikidata.org/entity/>

        SELECT ?localitem ?ext_id WHERE {{
          SERVICE <https://query.wikidata.org/sparql> {{
            ?wditem wwdt:{prop} ?ext_id .
            {filter_str}
          }}
          ?localitem wdt:P3 ?wditem
        }}
        """.format(prop=prop, filter_str=filter_str)

        results = wdi_core.WDItemEngine.execute_sparql_query(query, endpoint=self.sparql_endpoint_url)['results']['bindings']
        results = [{k: v['value'] for k, v in x.items()} for x in results]
        for r in results:
            r['localitem'] = r['localitem'].split('/')[-1]
        if not results:
            return None

        id_qid = defaultdict(set)
        for r in results:
            id_qid[r['ext_id']].add(r['localitem'])

        if return_as_set:
            return dict(id_qid)
        else:
            return {x['ext_id']: x['localitem'] for x in results}


import copy
import pprint
import unittest

import requests

from wikidataintegrator import wdi_core, wdi_fastrun
from wikidataintegrator.wdi_core import WDApiError

__author__ = 'Sebastian Burgstaller-Muehlbacher'
__license__ = 'AGPLv3'


class TestMediawikiApiCall(unittest.TestCase):
    def test_all(self):
        with self.assertRaises(WDApiError):
            wdi_core.WDItemEngine.mediawiki_api_call("GET", "http://www.wikidataaaaaaaaa.org",
                                                     max_retries=3, retry_after=1,
                                                     params={'format': 'json', 'action': 'wbgetentities', 'ids': 'Q42'})
        with self.assertRaises(requests.HTTPError):
            wdi_core.WDItemEngine.mediawiki_api_call("GET", "http://httpstat.us/400", max_retries=3, retry_after=1)

        wdi_core.WDItemEngine.mediawiki_api_call("GET", max_retries=3, retry_after=1,
                                                 params={'format': 'json', 'action': 'wbgetentities', 'ids': 'Q42'})


class TestDataType(unittest.TestCase):
    def test_wd_quantity(self):
        dt = wdi_core.WDQuantity(value='34', prop_nr='P43')

        dt_json = dt.get_json_representation()

        if not dt_json['mainsnak']['datatype'] == 'quantity':
            raise

        value = dt_json['mainsnak']['datavalue']

        if not value['value']['amount'] == '+34':
            raise

        if not value['value']['unit'] == '1':
            raise

        dt2 = wdi_core.WDQuantity(value='34', prop_nr='P43', upper_bound='35', lower_bound='33')

        value = dt2.get_json_representation()['mainsnak']['datavalue']

        if not value['value']['amount'] == '+34':
            raise

        if not value['value']['unit'] == '1':
            raise

        if not value['value']['upperBound'] == '+35':
            raise

        if not value['value']['lowerBound'] == '+33':
            raise

    def test_wd_geoshape(self):
        dt = wdi_core.WDGeoShape(value='Data:Inner_West_Light_Rail_stops.map', prop_nr='P43')

        dt_json = dt.get_json_representation()

        if not dt_json['mainsnak']['datatype'] == 'geo-shape':
            raise

        value = dt_json['mainsnak']['datavalue']

        if not value['value'] == 'Data:Inner_West_Light_Rail_stops.map':
            raise

        if not value['type'] == 'string':
            raise

    def test_wd_string(self):
        pass

    def test_live_item(self):
        wd_item = wdi_core.WDItemEngine(wd_item_id='Q423111')

        mass_statement = [x for x in wd_item.statements if x.get_prop_nr() == 'P2067'].pop()
        pprint.pprint(mass_statement.get_json_representation())

        if not mass_statement:
            raise

            # TODO: get json directly from the API and compare part to WDItemEngine


class TestFastRun(unittest.TestCase):
    """
    some basic tests for fastrun mode
    
    """

    def test_fast_run(self):
        qid = 'Q27552312'

        statements = [
            wdi_core.WDExternalID(value='P40095', prop_nr='P352'),
            wdi_core.WDExternalID(value='YER158C', prop_nr='P705')
        ]

        frc = wdi_fastrun.FastRunContainer(base_filter={'P352': '', 'P703': 'Q27510868'},
                                           base_data_type=wdi_core.WDBaseDataType, engine=wdi_core.WDItemEngine)

        fast_run_result = frc.write_required(data=statements)

        if fast_run_result:
            message = 'fastrun failed'
        else:
            message = 'successful fastrun'
        print(fast_run_result, message)

        # here, fastrun should succeed, if not, test failed
        # if fast_run_result:
        #    raise ValueError

    def test_fastrun_label(self):
        # tests fastrun label, description and aliases, and label in another language
        data = [wdi_core.WDExternalID('/m/02j71', 'P646')]
        fast_run_base_filter = {'P361': 'Q18589965'}
        item = wdi_core.WDItemEngine(wd_item_id="Q2", data=data, fast_run=True,
                                     fast_run_base_filter=fast_run_base_filter)

        frc = wdi_core.WDItemEngine.fast_run_store[0]
        frc.debug = True

        assert item.get_label('en') == "Earth"
        descr = item.get_description('en')
        assert len(descr) > 3
        aliases = item.get_aliases()
        assert "Terra" in aliases

        assert list(item.fast_run_container.get_language_data("Q2", 'en', 'label'))[0] == "Earth"
        assert item.fast_run_container.check_language_data("Q2", ['not the Earth'], 'en', 'label')
        assert "Terra" in item.get_aliases()
        assert "planet" in item.get_description()

        assert item.get_label("es") == "Tierra"

        item.set_description(descr)
        item.set_description("fghjkl")
        assert item.wd_json_representation['descriptions']['en'] == {'language': 'en', 'value': 'fghjkl'}
        item.set_label("Earth")
        item.set_label("xfgfdsg")
        assert item.wd_json_representation['labels']['en'] == {'language': 'en', 'value': 'xfgfdsg'}
        item.set_aliases(["fake alias"], append=True)
        assert {'language': 'en', 'value': 'fake alias'} in item.wd_json_representation['aliases']['en']

        # something thats empty (for now.., can change, so this just makes sure no exception is thrown)
        frc.check_language_data("Q2", ['Ewiase'], 'ak', 'label')
        frc.check_language_data("Q2", ['not Ewiase'], 'ak', 'label')
        frc.check_language_data("Q2", [''], 'ak', 'description')
        frc.check_language_data("Q2", [], 'ak', 'aliases')
        frc.check_language_data("Q2", ['sdf', 'sdd'], 'ak', 'aliases')

        item.get_label("ak")
        item.get_description("ak")
        item.get_aliases("ak")
        item.set_label("label", lang="ak")
        item.set_description("d", lang="ak")
        item.set_aliases(["a"], lang="ak", append=True)


def test_sitelinks():
    data = [wdi_core.WDItemID(value='Q12136', prop_nr='P31')]
    item = wdi_core.WDItemEngine(wd_item_id='Q622901', data=data)
    item.get_sitelink("enwiki")
    assert "enwiki" not in item.wd_json_representation['sitelinks']
    item.set_sitelink("enwiki", "something")
    assert item.get_sitelink("enwiki")['title'] == "something"
    assert "enwiki" in item.wd_json_representation['sitelinks']


def test_nositelinks():
    # this item doesn't and probably wont ever have any sitelinks (but who knows?? maybe one day..)
    data = [wdi_core.WDItemID(value='Q5', prop_nr='P31')]
    item = wdi_core.WDItemEngine(wd_item_id='Q27869338', data=data)
    item.get_sitelink("enwiki")
    assert "enwiki" not in item.wd_json_representation['sitelinks']
    item.set_sitelink("enwiki", "something")
    assert item.get_sitelink("enwiki")['title'] == "something"
    assert "enwiki" in item.wd_json_representation['sitelinks']


####
## tests for statement equality, with and without refs
####
def test_ref_equals():
    # statements are identical
    oldref = [wdi_core.WDExternalID(value='P58742', prop_nr='P352'),
              wdi_core.WDItemID(value='Q24784025', prop_nr='P527'),
              wdi_core.WDTime('+2001-12-31T12:01:13Z', prop_nr='P813')]
    olditem = wdi_core.WDItemID("Q123", "P123", references=[oldref])
    newitem = copy.deepcopy(olditem)
    assert olditem.equals(newitem, include_ref=False)
    assert olditem.equals(newitem, include_ref=True)

    # dates are a month apart
    newitem = copy.deepcopy(olditem)
    newitem.references[0][2] = wdi_core.WDTime('+2002-1-31T12:01:13Z', prop_nr='P813')
    assert olditem.equals(newitem, include_ref=False)
    assert not olditem.equals(newitem, include_ref=True)

    # multiple refs
    newitem = copy.deepcopy(olditem)
    newitem.references.append([wdi_core.WDExternalID(value='99999', prop_nr='P352')])
    assert olditem.equals(newitem, include_ref=False)
    assert not olditem.equals(newitem, include_ref=True)
    olditem.references.append([wdi_core.WDExternalID(value='99999', prop_nr='P352')])
    assert olditem.equals(newitem, include_ref=True)



"""
custom ref handler:
overwrite a reference if the `stated in` is in `old_stated_in`
otherwise keep it and append a new ref

"""


def update_release(olditem, newitem, old_stated_in=None, stated_in_pid='P248', retrieved_pid='P813'):
    if not old_stated_in:
        old_stated_in = set()

    def is_equal_not_retrieved_stated_in(oldref, newref):
        # Return True if the oldref == newref, NOT including any "retrieved" or stated in statements
        if len(oldref) != len(newref):
            return False
        oldref_minus = [x for x in oldref if x.get_prop_nr() not in {retrieved_pid, stated_in_pid}]
        newref_minus = [x for x in newref if x.get_prop_nr() not in {retrieved_pid, stated_in_pid}]
        if not all(x in oldref_minus for x in newref_minus):
            return False
        return True

    def ref_overwrite(oldref, newref):
        if get_stated_in_value(oldref) & old_stated_in:
            # the old ref has a stated in that is `old`, so should be removed
            return True
        else:
            return False

    def get_stated_in_value(ref):
        stated_in_value = set(["Q{}".format(x.get_value()) for x in ref if x.get_prop_nr() == stated_in_pid])
        return stated_in_value

    newrefs = newitem.references
    oldrefs = olditem.references

    found_mate = [False] * len(newrefs)
    for new_n, newref in enumerate(newrefs):
        for old_n, oldref in enumerate(oldrefs):
            if is_equal_not_retrieved_stated_in(oldref, newref):
                found_mate[new_n] = True
                if ref_overwrite(oldref, newref):
                    oldrefs[old_n] = newref
    for f_idx, f in enumerate(found_mate):
        if not f:
            oldrefs.append(newrefs[f_idx])

"""
# todo: this needs to be integrated into a test system on a separate test wikibase
# todo: need another test for adding a ref with a release on an existing statement with no old release
from wikidataintegrator import wdi_core, wdi_login, wdi_helpers, ref_handlers
from functools import partial

mediawiki_api_url = "http://localhost:7171/w/api.php"
sparql_endpoint_url = "http://localhost:7272/proxy/wdqs/bigdata/namespace/wdq/sparql"
h = wdi_helpers.WikibaseHelper(sparql_endpoint_url)
stated_in = h.get_pid("P248")
retrieved = h.get_pid("P813")
doid = h.get_pid("P699")
login = wdi_login.WDLogin("testbot", "password", mediawiki_api_url=mediawiki_api_url)

# set the statement on sandbox, along with 2 other refs
s = wdi_core.WDExternalID("C1234", "P31", references=[
    [wdi_core.WDItemID("Q53798", stated_in, is_reference=True),
     wdi_core.WDTime("+2018-01-01T00:00:00Z", retrieved, is_reference=True),
     wdi_core.WDExternalID("DOID:0060852", doid, is_reference=True)],
    [wdi_core.WDItemID("Q123", stated_in, is_reference=True)],
    [wdi_core.WDExternalID("123", doid, is_reference=True)],
])

item = wdi_core.WDItemEngine(wd_item_id="Q80983", data=[s],
                             mediawiki_api_url=mediawiki_api_url, sparql_endpoint_url=sparql_endpoint_url)
item.write(login)

# update the stated in
s = wdi_core.WDExternalID("C1234", "P31", references=[
    [wdi_core.WDItemID("Q45050", stated_in, is_reference=True),
     wdi_core.WDTime("+2019-02-01T00:00:00Z", retrieved, is_reference=True),
     wdi_core.WDExternalID("DOID:0060852", doid, is_reference=True)]])


ref_handler = partial(ref_handlers.update_release, old_stated_in={'Q53798'},
                      stated_in_pid=stated_in, retrieved_pid=retrieved)
item = wdi_core.WDItemEngine(wd_item_id="Q80983", data=[s],
                             mediawiki_api_url=mediawiki_api_url, sparql_endpoint_url=sparql_endpoint_url,
                             ref_handler=ref_handler, global_ref_mode="CUSTOM")
item.write(login)
"""


import unittest
from wikidataintegrator import wdi_core

__author__ = 'Sebastian Burgstaller-Muehlbacher'
__license__ = 'AGPLv3'


class TestWDSearch(unittest.TestCase):

    def test_wd_search(self):
        t = wdi_core.WDItemEngine.get_wd_search_results('rivaroxaban')

        print(t)
        print('Number of results: ', len(t))
        self.assertIsNot(len(t), 0)
